﻿using System;

namespace InABox.Core
{

    // We removed ISkipLoad because it seemed pointless.
    [UserTracking(typeof(User))]
    public class Document : Entity, IRemotable, IPersistent, ILicense<CoreLicense>
    {
        [TextBoxEditor]
        [EditorSequence(1)]
        public string FileName { get; set; }

        [CheckBoxEditor]
        [EditorSequence(2)]
        public bool Private { get; set; }

        [TimestampEditor(Editable = Editable.Hidden)]
        public DateTime TimeStamp { get; set; }

        [CodeEditor(Editable = Editable.Enabled)]
        public string CRC { get; set; }

        [NullEditor]
        public byte[] Data { get; set; }

        /*[DoNotSerialize]
        [DoNotPersist]
        [NullEditor]
        public string AsBase64
        {
            get => Data == null ? string.Empty : Convert.ToBase64String(Data);
            set => Data = Convert.FromBase64String(value);
        }*/

        [TimestampEditor(Editable = Editable.Hidden)]
        public override DateTime Created
        {
            get => base.Created;
            set => base.Created = value;
        }

        [TextBoxEditor(Editable = Editable.Hidden)]
        public override string CreatedBy
        {
            get => base.CreatedBy;
            set => base.CreatedBy = value;
        }

        protected override void Init()
        {
            base.Init();
            Data = new byte[] { };
        }

        public override string ToString()
        {
            return FileName;
        }
    }
}